package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.dao.RuleRepository;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.Rule;
import org.springframework.beans.factory.annotation.Autowired;

import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import javax.transaction.Transactional;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;


import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;

import java.util.HashSet;
import java.util.List;

import java.util.Set;
import java.util.stream.Collectors;


/**
 * @author Jakub Uhlarik
 */
@Service
public class GrammarService {

    public static class XmlGrammarHandler extends DefaultHandler {
        private Grammar newGrammar;
        private StringBuilder rightSide;
        private boolean oneOf = false;
        private Rule rule = null;
        private boolean item = false;

        @Override
        public void startDocument() {
            newGrammar = new Grammar(null, new HashSet<>(), null);
        }

        @Override
        public void characters(char[] ch, int start, int length) {
            if (rightSide != null && item) {
                rightSide.append(ch, start, length);
            }
        }

        @Override
        public void startElement(String uri, String lName, String qName, Attributes attr) {
            switch (qName) {
                case "grammar" -> {
                    String grammarUri = attr.getValue("xml:base");
                    if (grammarUri != null) {
                        newGrammar.setUri(grammarUri);
                    }
                    String rootLeft = attr.getValue("root");
                    if (rootLeft != null) {
                        newGrammar.getRules().add(new Rule(newGrammar, rootLeft, "", true));
                    }
                }
                case "rule" -> {
                    rule = new Rule(newGrammar, attr.getValue("id"), "", false);
                    rightSide = new StringBuilder();
                }
                case "item" -> {
                    item = true;
                    if (oneOf && rightSide.length() != 0) {
                        rightSide.append("|");
                    }
                }
                case "one-of" -> oneOf = true;
                case "ruleref" -> {

                    String spRule = attr.getValue("special");
                    if (spRule != null) {
                        rightSide.append("<").append(spRule).append(">");
                    }
                    String ruleRef = attr.getValue("uri");
                    if (ruleRef != null) {
                        if (ruleRef.startsWith("#")) {
                            ruleRef = ruleRef.substring(1);
                        }
                        rightSide.append("<").append(ruleRef).append(">");
                    }

                }
            }
        }

        @Override
        public void endElement(String uri, String lName, String qName) {
            switch (qName) {
                case "rule" -> {
                    rule.setRightSide(rightSide.toString());
                    newGrammar.getRules().add(rule);
                    rule = null;
                }
                case "item" -> item = false;
                case "one-of" -> oneOf = false;
            }
        }


    }

    private final GrammarRepository grammarRepository;
    private final RuleRepository ruleRepository;

    private final RuleService ruleService;

    @Autowired
    public GrammarService(GrammarRepository grammarRepository,
                          RuleRepository ruleRepository, RuleService ruleService) {
        this.grammarRepository = grammarRepository;
        this.ruleRepository = ruleRepository;
        this.ruleService = ruleService;
    }

    public List<Grammar> getGrammars() {
        return grammarRepository.findAll();
    }


    public void deleteAll() {
        ruleService.deleteRules();
        grammarRepository.deleteAll();
    }


    @Transactional
    public void deleteById(Long id) {
        Grammar grammar = getGrammarById(id);
        ruleRepository.deleteAll(grammar.getRules());
        grammarRepository.deleteById(id);
    }


    @Transactional
    public Grammar addGrammar(Grammar grammar) {
        if (grammar.getId() != null && grammarRepository.existsById(grammar.getId())) {
            grammar.setId(null);
        }
        if (grammar.getUri() == null) {
            grammar.setUri("null");
        }

        Boolean exists_uri = grammarRepository.existsByUri(grammar.getUri());
        Rule rootRule = grammar.getRootRule();
        Set<Rule> rules = grammar.getRules();
        if (rules == null) {
            rules = new HashSet<>();
        }

        grammar.setRules(null);
        grammar.setRootRule(null);
        grammar = grammarRepository.save(grammar);
        //uri modification
        if (exists_uri) {
            grammar.setUri(generateUniqueUri(grammar.getUri(), grammar.getId()));
        }
        if (rootRule != null) {
            rootRule.setRoot(true);
            rules.remove(rootRule);
            ruleService.addRule(rootRule, grammar);
        }
        for (Rule rule : rules) {
            //@todo split preprocess
            ruleService.addRule(rule, grammar);
        }

        if (grammar.getRules() == null) {
            grammar.setRules(new HashSet<>());
        }


        return grammar;
    }

    public Grammar getGrammarById(Long id) {
        return grammarRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "grammar id: " + id + " not found"));
    }

    @Transactional
    public Grammar duplicateGrammar(Long id) {
        Grammar grammar = getGrammarById(id);
        Grammar newGrammar = new Grammar();
        newGrammar.setUri(grammar.getUri());
        Rule oldRule = grammar.getRootRule();
        if (oldRule != null) {
            newGrammar.setRootRule(new Rule(oldRule));
        }
        Set<Rule> rules = grammar.getRules();
        Set<Rule> newRules = new HashSet<>();
        for (Rule rule : rules) {
            newRules.add(new Rule(rule));
        }
        newGrammar.setRules(newRules);
        newGrammar = addGrammar(newGrammar);

        return newGrammar;
    }

    public Set<Grammar> getGrammarByRootRule(String rootSymbol) {
        List<Rule> rulesWithSymbol = ruleRepository.findByLeftSymbol(rootSymbol);
        return rulesWithSymbol.stream().filter((Rule::getRoot)).map((Rule::getGrammar)).collect(Collectors.toSet());

    }

    public Set<Grammar> getGrammarsByUriStart(String uri) {
        return grammarRepository.findByUriStartingWith(uri);
    }


    private void printRightSide(PrintStream bf, String rightSide) {
        if (rightSide.contains("<") && rightSide.contains(">")) {

            int leftindex = 0;
            int rightindex = 0;

            int begin = 0;
            while (leftindex >= 0 && rightindex >= 0) {
                leftindex = rightSide.indexOf('<', rightindex);
                rightindex = rightSide.indexOf('>', leftindex);
                //between two rules or begin
                if (leftindex - begin > 0) {

                    bf.print(rightSide.substring(begin, leftindex));
                }
                //after rules
                if (leftindex == -1 && rightSide.length() > begin) {
                    bf.print(rightSide.substring(begin));
                }
                if (leftindex >= 0 && rightindex >= 0) {
                    String ruleInside = rightSide.substring(leftindex + 1, rightindex);
                    if (ruleInside.equals("NULL")) {
                        bf.print("<ruleref special=\"NULL\"/>");
                    } else if (ruleInside.equals("VOID")) {
                        bf.print("<ruleref special=\"VOID\"/>");
                    } else {
                        bf.print("<ruleref uri=\"#" + ruleInside + "\"/>");
                    }

                }
                begin = rightindex + 1;

            }
            return;
        }
        if (!rightSide.contains("<") && !rightSide.contains(">")) {
            bf.println(rightSide);
            return;
        }
        throw new ResponseStatusException(HttpStatus.EXPECTATION_FAILED, "bad rule format");
    }

    private void writeRule(PrintStream bf, Rule rule, boolean xml) {
        if (!xml) {
            if (rule.getRoot()) {
                bf.println("//root;");
            }
            bf.println(rule.getLeftSymbol() + " = " + rule.getRightSide() + ";");
            return;
        }
        bf.print("<rule id=\"");
        bf.println(nameFromLeftSymbol(rule) + "\">");
        String rightSide = rule.getRightSide();
        if (rightSide.contains("|")) {
            String[] splitString = rightSide.split("\\|");
            bf.println("<one-of>");
            for (String partString : splitString) {
                bf.print("<item>");
                printRightSide(bf, partString);
                bf.println("</item>");
            }
            bf.println("</one-of>");
        } else {
            bf.print("<item>");
            printRightSide(bf, rightSide);
            bf.println("</item>");
        }
        bf.println("</rule>");
    }

    private String nameFromLeftSymbol(Rule rule) {
        String oldLeftSymbol = rule.getLeftSymbol();
        return rule.getLeftSymbol().substring(1, oldLeftSymbol.length() - 1);
    }

    private void writeHeader(PrintStream ps, Grammar grammar, boolean xml) {
        //jsgf
        if (!xml) {
            ps.println("#JSGF v 1.0;");
            ps.println("grammar " + grammar.getUri() + ";\n");
            return;
        }
        //xml
        ps.println("<?xml version=\"1.0\"?>");
        ps.println("<grammar version=\"1.0\"");
        Rule rootRule;
        String uri;
        if ((uri = grammar.getUri()) != null) {
            ps.println("xml:base=\"" + uri + "\"");
        }
        if ((rootRule = grammar.getRootRule()) != null) {
            ps.println("root=\"" + nameFromLeftSymbol(rootRule) + "\"");
        }

        ps.println(" xmlns=\"http://www.w3.org/2001/06/grammar\">");
    }

    public void fileFromGrammar(Long id, boolean xml) throws IOException {
        Grammar grammar = getGrammarById(id);
        String filePath = "./grammars/" + id + (xml ? ".xml" : ".jsgf");
        File file = new File(filePath);

        boolean exists = file.exists();
        boolean created = file.createNewFile();
        if (!exists && !created) {
            throw new RuntimeException("file error");
        }
        try (PrintStream ps = new PrintStream(new FileOutputStream(file))) {
            writeHeader(ps, grammar, xml);

            Rule rootRule;
            if (null != (rootRule = grammar.getRootRule())) {
                writeRule(ps, rootRule, xml);
            }

            for (Rule rule : grammar.getRules()) {
                if (!rule.getRoot()) {
                    writeRule(ps, rule, xml);
                }
            }
            if (xml) {
                ps.println("</grammar>");
            }
        }
    }


    public Grammar addGrammarFromFile(String file, boolean xml) {
        Grammar grammar = new Grammar();


        if (xml) {
            try {
                SAXParserFactory factory = SAXParserFactory.newInstance();
                SAXParser saxParser = factory.newSAXParser();

                XmlGrammarHandler handler = new XmlGrammarHandler();
                InputSource is = new InputSource(new StringReader(file));
                saxParser.parse(is, handler);

                grammar = handler.newGrammar;

            } catch (ParserConfigurationException | SAXException | IOException e) {
                throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "parsing error", e);

            }


        } else {
            boolean comment = false;
            boolean complexCommentNow;
            String[] lines = file.split("[;\n]");
            String uncommentedLine = null;
            int i = 1;
            boolean root = false;
            boolean setRoot = false;
            for (String line : lines) {
                complexCommentNow = false;
                if (setRoot) {
                    root = true;
                    setRoot = false;
                }
                line = line.trim();
                if (line.contains("/*")) {
                    String[] splitByComment = line.split("/\\*");
                    uncommentedLine = splitByComment[0];
                    comment = true;
                    complexCommentNow = true;
                }
                if (line.contains("*/")) {
                    if (comment) {
                        comment = false;
                    } else {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "comment without start");
                    }
                    String[] splitByComment = line.split("\\*/");
                    line = splitByComment[1];
                    if (complexCommentNow) {
                        line = uncommentedLine + line;
                    }


                }
                if (line.contains("//")) {
                    String[] splitByComment = line.split("//");
                    line = splitByComment[1];
                    if (line.equals("root")) {
                        setRoot = true;
                    }
                }
                if (line.contains("#JSGF")) {
                    i++;
                    continue;
                }
                if (line.contains("grammar ")) {
                    if (grammar.getUri() != null) {
                        throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "duplicate grammar name declaration;line=" + i);
                    }
                    String[] splitLine = line.split(" ");
                    grammar.setUri(splitLine[1]);
                }
                if (line.contains("=")) {
                    //rule
                    String[] splitLine = line.split("=");
                    String leftSymbol = ruleService.stringEncapsulation(splitLine[0].trim());

                    Rule rule = new Rule(grammar, leftSymbol, splitLine[1].trim(), root);
                    root = false;
                    if (grammar.getRules() == null) {
                        grammar.setRules(new HashSet<>());
                    }
                    grammar.getRules().add(rule);
                }
                i++;
            }

        }
        return addGrammar(grammar);
    }

    private String generateUniqueUri(String uri, Long id) {
        String result = uri + id;
        while (grammarRepository.existsByUri(result)) {
            result = uri + id;
            result += "." + (int) (Math.random() * 100000);
        }
        return result;
    }

    @Transactional
    public Grammar changeUri(Long id, Grammar grammar) {
        Grammar oldGrammar = this.getGrammarById(id);
        if (grammar.getUri() != null) {
            oldGrammar.setUri(null);

            if (grammarRepository.existsByUri(grammar.getUri())) {
                oldGrammar.setUri(generateUniqueUri(grammar.getUri(), oldGrammar.getId()));
                return oldGrammar;
            }
            oldGrammar.setUri(grammar.getUri());
        }
        return oldGrammar;
    }
}
