package com.jakubu9333.grammarManager.service;

import com.jakubu9333.grammarManager.dao.GrammarRepository;
import com.jakubu9333.grammarManager.dao.RuleRepository;
import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.Rule;
import org.javatuples.Pair;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.stereotype.Service;
import org.springframework.web.server.ResponseStatusException;

import javax.transaction.Transactional;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * @author Jakub Uhlarik
 */
@Service
public class RuleService {

    private final RuleRepository ruleRepository;
    private final GrammarRepository grammarRepository;

    @Autowired
    public RuleService(RuleRepository ruleRepository, GrammarRepository grammarRepository) {
        this.ruleRepository = ruleRepository;
        this.grammarRepository = grammarRepository;

    }

    public List<Rule> getAllRules() {
        return ruleRepository.findAll();
    }

    private Rule saveRule(Rule rule) {
        return ruleRepository.save(rule);
    }

    public String stringEncapsulation(String string) {
        if (string == null) {
            return null;
        }
        StringBuilder stringBuilder = new StringBuilder();
        if (!string.startsWith("<")) {
            stringBuilder.append('<');
        }
        stringBuilder.append(string);
        if (!string.endsWith(">")) {
            stringBuilder.append('>');
        }
        return stringBuilder.toString();

    }

    private String[] splitRightSide(String string) {
        if (string == null) {
            return new String[]{""};
        }
        if (string.contains("|")) {
            return string.split("\\|");
        }
        return new String[]{string};
    }

    private String normalBrackets(String string) {
        StringBuilder sb = new StringBuilder(string);

        int leftIndex;

        int rightIndex = -1;
        while ((leftIndex = sb.indexOf("(")) != -1) {
            int brackets1 = 0;
            for (int i = leftIndex; i < sb.length(); i++) {
                if (sb.charAt(i) == '(') {
                    brackets1++;
                }
                if (sb.charAt(i) == ')') {
                    brackets1--;
                    if (brackets1 == 0) {
                        rightIndex = i;
                        break;
                    }
                }

            }
            if (rightIndex == -1) {
                break;
            }

            String inside = sb.substring(leftIndex + 1, rightIndex);
            int outsideLeft = 0;
            int outSideRight = 0;
            for (int i = leftIndex; i >= 0; i--) {
                outsideLeft = i;
                if (sb.charAt(i) == '|') {
                    break;
                }
            }
            int brackets = 0;
            for (int i = rightIndex; i < sb.length(); i++) {
                outSideRight = i;

                if (sb.charAt(i) == '(') {
                    brackets++;
                }
                if (sb.charAt(i) == ')' && i != rightIndex) {
                    brackets--;
                }
                if (sb.charAt(i) == '|' && brackets == 0) {
                    break;
                }
            }
            if (outsideLeft != 0 && leftIndex != 0) {
                outsideLeft++;
            }

            String leftOutside = sb.substring(outsideLeft, leftIndex);
            String rightOutside = sb.substring(rightIndex + 1, outSideRight + 1);
            sb.replace(outsideLeft, outSideRight + 1, "|");
            for (String stringInSplit : splitRightSide(inside)) {
                sb.append("|").append(leftOutside).append(stringInSplit).append(rightOutside);
            }


        }
        return sb.toString();
    }

    private String straightBrackets(String string) {
        StringBuilder sb = new StringBuilder(string);

        int leftIndex;

        int rightIndex;
        while ((leftIndex = sb.indexOf("[")) != -1 && (rightIndex = sb.indexOf("]")) != -1) {

            String inside = sb.substring(leftIndex + 1, rightIndex);
            sb.replace(leftIndex, rightIndex + 1, "(" + inside + "|<NULL>)");

        }
        return sb.toString();

    }

    private List<Pair<Integer, Integer>> getIgnoreListCheckRuleRef(String addedRightSide) {
        List<Pair<Integer, Integer>> result = new ArrayList<>();
        Integer start = null;
        Integer end = null;
        Integer startAngleBracket = null;
        Integer endAngleBracket = null;
        for (int i = 0; i < addedRightSide.length(); i++) {
            boolean slashNotBefore =((i == 0 || addedRightSide.charAt(i - 1) != '\\') || (i > 1 && addedRightSide.charAt(i - 1) == '\\' && addedRightSide.charAt(i - 2) == '\\'));
            if (addedRightSide.charAt(i) == '"' &&  slashNotBefore) {
                if (start == null) {
                    end = null;
                    start = i;
                } else {
                    result.add(new Pair<>(start, i));
                    start = null;
                    end = i;
                }
            }
            if (addedRightSide.charAt(i) == '<' && slashNotBefore) {
                endAngleBracket = null;
                startAngleBracket = i;
            }
            if (addedRightSide.charAt(i) == '>' && slashNotBefore) {
                if (startAngleBracket == null) {
                    throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "ending angle bracket without starting bracket");
                }
                endAngleBracket = i;
                result.add(new Pair<>(startAngleBracket, endAngleBracket));
                startAngleBracket = null;

            }
        }
        if (end == null && start != null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "last quote not ended");
        }
        if (endAngleBracket == null && startAngleBracket != null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "ruleref not ended");
        }
        return result;
    }

    private String newRightSide(Rule oldRule, Rule newRule) {
        String oldRightSide = oldRule.getRightSide() == null ? "" : oldRule.getRightSide();
        String addedRightSide = newRule.getRightSide() == null ? "<NULL>" : newRule.getRightSide();
        String[] oldRightSideSplit = splitRightSide(oldRightSide);
        List<Pair<Integer, Integer>> ignoreList = getIgnoreListCheckRuleRef(addedRightSide);
        boolean straightBrackets = (addedRightSide.contains("[")) && (addedRightSide.contains("]"));
        if (straightBrackets) {
            addedRightSide = straightBrackets(addedRightSide);
        }
        boolean normalBrackets = (addedRightSide.contains("(")) && (addedRightSide.contains(")"));
        if (normalBrackets) {
            addedRightSide = normalBrackets(addedRightSide);
        }
        String[] newRightSideSplit = splitRightSide(addedRightSide);
        Set<String> newRightSideParts = Arrays.stream(newRightSideSplit).filter((elem) -> elem.length() != 0).map(String::trim).collect(Collectors.toSet());
        newRightSideParts.removeAll(Arrays.stream(oldRightSideSplit).filter((elem) -> elem.length() != 0).collect(Collectors.toSet()));


        StringBuilder newRightSide = new StringBuilder(oldRightSide);
        for (String newRightSidePart : newRightSideParts) {
            String middle = newRightSide.length() == 0 ? "" : "|";
            newRightSide.append(middle).append(newRightSidePart);
        }
        if (newRightSide.length() == 0) {
            newRightSide.append("<NULL>");
        }
        return newRightSide.toString();
    }


    @Transactional
    public Pair<Rule, Boolean> rulePreProcess(Rule rule, Grammar grammar) {
        String leftSymbol = rule.getLeftSymbol();
        if (leftSymbol == null) {
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "no left symbol");
        }
        if (leftSymbol.trim().length() == 0) {
            rule.setLeftSymbol("empty");
        }
        boolean isRoot = rule.getRoot();

        //checks if any in this grammar rule have same left side
        //@todo change multiple add to make able splitting pre process
        List<Rule> oldRules = getByLeftSide(rule.getLeftSymbol(), grammar.getId());
        if (!oldRules.isEmpty() && !Objects.equals(oldRules.get(0).getId(), rule.getId())) {
            Rule oldRule = oldRules.get(0);
            oldRule.setRightSide(newRightSide(oldRule, rule));
            if (grammar.getRootRule() == null && isRoot) {
                oldRule.setRoot(true);
                grammar.setRootRule(oldRule);
            }
            return new Pair<>(oldRule, false);
        } else {
            // parse right side of new rule
            rule.setRightSide(newRightSide(new Rule(), rule));
        }


        rule.setLeftSymbol(stringEncapsulation(rule.getLeftSymbol()));
        return new Pair<>(rule, true);
    }

    @Transactional
    public Pair<Rule, Boolean> addRule(Rule rule, Grammar grammar) {
        if (rule.getId() != null && ruleRepository.existsById(rule.getId())) {
            rule.setId(null);
        }
        if (rule.getRoot() == null) {
            rule.setRoot(false);
        }
        Pair<Rule, Boolean> result = rulePreProcess(rule, grammar);
        rule = result.getValue0();
        //if no new rule will be created
        if (!result.getValue1()) {
            return result;
        }

        if (grammar.getRootRule() != null && rule.getRoot()) {
            rule.setRoot(false);
        }

        rule = saveRule(rule);

        Set<Rule> rules = grammar.getRules();
        if (rule.getRoot()) {
            grammar.setRootRule(rule);
        }
        rules.add(rule);
        grammar.setRules(rules);
        rule.setGrammar(grammar);
        return new Pair<>(rule, true);
    }

    @Transactional
    public void deleteRules() {
        grammarRepository.findAll().forEach(((grammar) -> {
            grammar.setRules(null);
            grammar.setRootRule(null);
        }));
        ruleRepository.deleteAll();
    }

    @Transactional
    public void deleteById(Long id) {
        Rule rule = ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND, "rule id: " + id + " not found"));
        Grammar rulesGrammar = rule.getGrammar();
        rulesGrammar.setRules(rulesGrammar.getRules().stream().filter((rule1 -> rule1 != rule)).collect(Collectors.toSet()));
        if (rule.equals(rulesGrammar.getRootRule())) {
            rulesGrammar.setRootRule(null);
        }
        ruleRepository.delete(rule);
    }


    public List<Rule> getByLeftSide(String leftSymbol) {
        leftSymbol = stringEncapsulation(leftSymbol);
        return ruleRepository.findByLeftSymbol(leftSymbol);
    }

    public List<Rule> getByLeftSide(String leftSymbol, Long grammarId) {
        leftSymbol = stringEncapsulation(leftSymbol);
        return ruleRepository.findByLeftSymbolAndGrammar_Id(leftSymbol, grammarId);
    }

    public Rule getById(Long id) {
        return ruleRepository.findById(id).orElseThrow(() -> new ResponseStatusException(HttpStatus.NOT_FOUND));
    }

    @Transactional
    public Rule updateRule(Long id, Rule rule) {
        Rule oldRule = this.getById(id);
        if (rule.getRoot() != null && oldRule.getRoot() != rule.getRoot()) {
            oldRule = changeRootStatus(id);
        }
        boolean changes = false;
        if (rule.getLeftSymbol() != null && rule.getLeftSymbol().length() != 0) {
            oldRule.setLeftSymbol(rule.getLeftSymbol());
            changes = true;
        }
        if (rule.getRightSide() != null && rule.getRightSide().length() != 0) {
            oldRule.setRightSide(rule.getRightSide());
            changes = true;
        }
        if (changes) {
            return this.rulePreProcess(oldRule, oldRule.getGrammar()).getValue0();
        }
        return oldRule;
    }

    @Transactional
    public Rule changeRootStatus(Long id) {
        Rule rule = this.getById(id);
        Grammar grammar = rule.getGrammar();
        if (rule.getRoot()) {
            grammar.setRootRule(null);
            rule.setRoot(false);
            return rule;
        }
        if (grammar.getRootRule() != null) {
            //return;
            throw new ResponseStatusException(HttpStatus.BAD_REQUEST, "grammar have root terminal");
        }
        grammar.setRootRule(rule);
        rule.setRoot(true);
        return rule;

    }

    public Pair<List<Rule>, Boolean> addRuleMultiple(List<Rule> rules, Grammar grammar) {
        Boolean change = false;
        List<Rule> result = new ArrayList<>();
        for (Rule rule : rules) {
            Pair<Rule, Boolean> partResult = addRule(rule, grammar);
            change |= partResult.getValue1();
            result.add(partResult.getValue0());
        }
        return new Pair<>(result, change);

    }
}
