package com.jakubu9333.grammarManager;


import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.Rule;
import com.jakubu9333.grammarManager.service.GrammarService;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.web.server.ResponseStatusException;

import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * @author Jakub Uhlarik
 */


@SpringBootTest
public class GrammarServiceTest {
    @Autowired
    private GrammarService grammarService;

    @Test
    public void testAddBasicGrammars() {
        Grammar grammarEmpty = new Grammar();
        Grammar grammarJustRootRule = new Grammar("xdd", null, new Rule(null, "<S>", "pp", false));
        Set<Rule> oldrules = new HashSet<>();
        oldrules.add(new Rule(null, "xd", "xd", true));
        Grammar grammarRootRule = new Grammar("xddd", oldrules, null);
        grammarRootRule = grammarService.addGrammar(grammarRootRule);
        assert grammarRootRule.getRootRule() != null;
        assert grammarRootRule.getRootRule().getLeftSymbol().equals("<xd>");
        grammarEmpty = grammarService.addGrammar(grammarEmpty);
        grammarJustRootRule = grammarService.addGrammar(grammarJustRootRule);
        assert null != grammarEmpty.getId();
        assert grammarEmpty.getRules() != null;
        assert grammarEmpty.getUri().equals("null");
        assert grammarJustRootRule.getId() != null;
        assert grammarJustRootRule.getUri().equals("xdd");
        Set<Rule> rules = grammarJustRootRule.getRules();
        assert rules != null;
        assert rules.size() == 1;
        Rule rootRule = grammarJustRootRule.getRootRule();
        assert rootRule.getRoot();
        assert rootRule.getLeftSymbol().equals("<S>");
        assert rootRule.getRightSide().equals("pp");
        assert rootRule.getId() != null;
        assert rootRule.getGrammar().equals(grammarJustRootRule);

    }

    @Test
    public void testSameUriGrammar() {
        grammarService.deleteAll();
        Grammar grammar1 = new Grammar("demoUri", null, null);
        Grammar grammar2 = new Grammar("demoUri", null, null);
        grammar1 = grammarService.addGrammar(grammar1);
        grammar2 = grammarService.addGrammar(grammar2);

        assert grammar1.getUri().equals("demoUri");
        assert grammar2.getUri().matches("demoUri[0-9]+");
        Long id = grammar2.getId();

        Grammar grammar3 = grammarService.addGrammar(new Grammar("demoUri" + (id + 2), null, null));
        assert grammar3.getId() == id + 1;
        Grammar grammar4 = grammarService.addGrammar(new Grammar("demoUri", null, null));
        assert grammar4.getId() == id + 2;
        assert grammar4.getUri().matches("demoUri[0-9]+.[0-9]+");
    }

    @Test
    public void testWithRules() {
        Set<Rule> differentRules = new HashSet<>();

        Rule rule1 = new Rule(null, "<p>", "pp", false);
        Rule rule2 = new Rule(null, "<g>", "gg", false);
        Rule rule3 = new Rule(null, "<r>", "rr", false);
        differentRules.add(rule1);
        differentRules.add(rule2);
        differentRules.add(rule3);
        Grammar grammarDifRules = new Grammar("difRules", differentRules, null);
        grammarDifRules = grammarService.addGrammar(grammarDifRules);
        assert grammarDifRules.getRules().equals(differentRules);
        for (Rule rule : grammarDifRules.getRules()) {
            assert rule.getGrammar().equals(grammarDifRules);
        }
    }

    @Test
    public void testSameRules() {
        Set<Rule> sameLeftRules = new HashSet<>();
        Rule rule1 = new Rule(null, "<p>", "pp", false);
        Rule rule2 = new Rule(null, "p", "prst", false);
        Rule rule3 = new Rule(null, "<p>", "p", false);
        Rule rule4 = new Rule(null, "<p>", "pp|pog", false);
        sameLeftRules.add(rule1);
        sameLeftRules.add(rule2);
        sameLeftRules.add(rule3);
        sameLeftRules.add(rule4);
        Grammar grammarSameRules = new Grammar("sameRules", sameLeftRules, null);
        grammarSameRules = grammarService.addGrammar(grammarSameRules);
        assert grammarSameRules.getRules().size() == 1;
        Rule processedRule = grammarSameRules.getRules().stream().toList().get(0);
        assert processedRule.getGrammar().equals(grammarSameRules);
        assert processedRule.getLeftSymbol().equals("<p>");
        String[] splitString = processedRule.getRightSide().split("\\|");
        int pp = 0;
        int prst = 0;
        int p = 0;
        int pog = 0;
        int def = 0;
        for (String smallStr : splitString) {
            switch (smallStr) {
                case "pp" -> pp++;
                case "p" -> p++;
                case "prst" -> prst++;
                case "pog" -> pog++;
                default -> def++;
            }

        }
        assert (pp == 1) && pog == 1 && prst == 1 && p == 1 && def == 0;


    }

    @Test
    public void testRootRuleWithRules() {
        Set<Rule> rulesRootRule = new HashSet<>();
        Rule rule1 = new Rule(null, "<p>", "pp", true);
        Rule rule2 = new Rule(null, "<g>", "pp", false);
        Rule rule3 = new Rule(null, "<s>", "pp", false);
        rulesRootRule.add(rule1);
        rulesRootRule.add(rule2);
        rulesRootRule.add(rule3);
        boolean exp = false;
        Grammar rootRuleGrammar = new Grammar("root", rulesRootRule, rule2);
        try {
            rootRuleGrammar = grammarService.addGrammar(rootRuleGrammar);
        } catch (ResponseStatusException e) {
            exp = true;
        }
        assert !exp;
        assert rootRuleGrammar.getRootRule().getLeftSymbol().equals("<g>");

    }

    @Test
    public void testAllGetDelete() {
        grammarService.deleteAll();
        Grammar grammar1 = new Grammar("1", null, null);
        Grammar grammar2 = new Grammar("2", null, null);
        Grammar grammar3 = new Grammar("3", null, null);
        Grammar grammar4 = new Grammar("4", null, null);
        grammarService.addGrammar(grammar1);
        grammarService.addGrammar(grammar2);
        grammarService.addGrammar(grammar3);
        grammarService.addGrammar(grammar4);
        List<Grammar> grammars = grammarService.getGrammars();
        assert grammars.size() == 4;
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        for (Grammar iterGram : grammars) {
            switch (iterGram.getUri()) {
                case "1" -> one++;
                case "2" -> two++;
                case "3" -> three++;
                case "4" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 0;
        grammarService.deleteAll();
        List<Grammar> grammars2 = grammarService.getGrammars();
        assert grammars2.size() == 0;
    }

    @Test
    public void testChangeUri() {
        grammarService.deleteAll();
        Grammar grammar1 = new Grammar("1x", null, null);
        Grammar grammar2 = new Grammar("2x", null, null);
        grammar1 = grammarService.addGrammar(grammar1);
        grammar2 = grammarService.addGrammar(grammar2);
        Long id = grammar2.getId();
        Grammar changeGr = new Grammar("xd", null, null);
        grammar2 = grammarService.changeUri(id, changeGr);
        assert Objects.equals(grammar2.getId(), id);
        assert grammar2.getUri().equals("xd");
        grammar1 = grammarService.changeUri(grammar1.getId(), changeGr);
        assert grammar1.getUri().equals("xd" + grammar1.getId());
        Grammar grammar3 = new Grammar("3x", null, null);
        Grammar changeGr2 = new Grammar("3x", null, null);
        grammar3 = grammarService.addGrammar(grammar3);
        changeGr2 = grammarService.changeUri(grammar3.getId(), changeGr2);
        assert changeGr2.getUri().equals("3x");
    }

    @Test
    public void testGetIdDeleteId() {
        grammarService.deleteAll();
        Grammar grammar1 = new Grammar("1x", null, null);
        Grammar grammar2 = new Grammar("2x", null, null);
        Grammar grammar3 = new Grammar("3x", null, null);
        Grammar grammar4 = new Grammar("4x", null, null);
        grammar1 = grammarService.addGrammar(grammar1);
        grammar2 = grammarService.addGrammar(grammar2);
        grammarService.addGrammar(grammar3);
        grammarService.addGrammar(grammar4);
        assert grammarService.getGrammarById(grammar1.getId()).getUri().equals("1x");
        grammarService.deleteById(grammar2.getId());
        List<Grammar> grammars = grammarService.getGrammars();
        assert grammars.size() == 3;
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        for (Grammar iterGram : grammars) {
            switch (iterGram.getUri()) {
                case "1x" -> one++;
                case "2x" -> two++;
                case "3x" -> three++;
                case "4x" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 0 && three == 1 && four == 1 && five == 0;
    }

    @Test
    public void testSpecialRootRuleGetters() {
        grammarService.deleteAll();
        Rule rule1 = new Rule(null, "<S>", "", true);
        Rule rule2 = new Rule(null, "<S>", "", true);
        Rule rule3 = new Rule(null, "<p>", "", true);
        Rule rule4 = new Rule(null, "<S>", "", false);
        Grammar grammar = new Grammar("g1", null, rule1);
        Grammar grammar2 = new Grammar("g2", null, rule2);
        Grammar grammar3 = new Grammar("g3", null, rule3);
        Set<Rule> rules = new HashSet<>();
        rules.add(rule4);
        Grammar grammar4 = new Grammar("g4", rules, null);
        grammarService.addGrammar(grammar);
        grammarService.addGrammar(grammar2);
        grammarService.addGrammar(grammar3);
        grammarService.addGrammar(grammar4);
        Set<Grammar> grammars = grammarService.getGrammarByRootRule("<S>");
        assert grammars.size() == 2;
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        for (Grammar iterGram : grammars) {
            switch (iterGram.getUri()) {
                case "g1" -> one++;
                case "g2" -> two++;
                case "g3" -> three++;
                case "g4" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 1 && three == 0 && four == 0 && five == 0;

    }

    @Test
    public void testUriStartUri() {
        grammarService.deleteAll();
        Grammar grammar = new Grammar("go1", null, null);
        Grammar grammar2 = new Grammar("go2", null, null);
        Grammar grammar3 = new Grammar("gl3", null, null);
        Grammar grammar4 = new Grammar("gofsdfjdsfpsdfosddslfksflk3", null, null);
        grammarService.addGrammar(grammar);
        grammarService.addGrammar(grammar2);
        grammarService.addGrammar(grammar3);
        grammarService.addGrammar(grammar4);
        Set<Grammar> grammars = grammarService.getGrammarsByUriStart("go");
        assert grammars.size() == 3;
        for (Grammar iterGram : grammars) {
            assert iterGram.getUri().startsWith("go");
        }
        assert grammarService.getGrammarsByUriStart("fsdfsdf").size() == 0;
    }
}
