package com.jakubu9333.grammarManager;

import com.jakubu9333.grammarManager.model.Grammar;
import com.jakubu9333.grammarManager.model.Rule;
import com.jakubu9333.grammarManager.service.GrammarService;
import com.jakubu9333.grammarManager.service.RuleService;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.web.server.ResponseStatusException;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Set;

/**
 * @author Jakub Uhlarik
 */
@SpringBootTest
public class RuleServiceTest {

    @Autowired
    private RuleService ruleService;
    @Autowired
    private GrammarService grammarService;

    @Test
    public void basicRulesAdd() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "h", "xd", false);
        Rule rule1 = new Rule(null, "g", "xd", true);
        rule = ruleService.addRule(rule, grammar).getValue0();
        ruleService.addRule(rule1, grammar);
        assert grammar.getRootRule().getLeftSymbol().equals("<g>");
        assert rule.getId() != null;
        Set<Rule> rules = grammar.getRules();
        assert rules.size() == 2;
        int one = 0;
        int two = 0;
        int three = 0;
        for (Rule smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "<h>" -> one++;
                case "<g>" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;
    }

    @Test
    public void sameLeftSideRule() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "g", "xd", false);
        Rule rule1 = new Rule(null, "<g>", "xd|good", true);
        Rule rule2 = new Rule(null, "g>", "xd|bad", true);
        Rule rule3 = new Rule(null, "d", "xd|smile", false);
        ruleService.addRule(rule, grammar);
        ruleService.addRule(rule1, grammar);
        rule = ruleService.addRule(rule2, grammar).getValue0();
        ruleService.addRule(rule3, grammar);
        int xd = 0;
        int good = 0;
        int smile = 0;
        int bad = 0;
        int def = 0;
        assert grammar.getRules().size() == 2;
        for (String leftPart : rule.getRightSide().split("\\|")) {
            switch (leftPart) {
                case "xd" -> xd++;
                case "good" -> good++;
                case "smile" -> smile++;
                case "bad" -> bad++;
                default -> def++;
            }
        }
        assert xd == 1 && good == 1 && smile == 0 && bad == 1 && def == 0;
    }

    @Test
    public void testEmptyAdd() {
        grammarService.deleteAll();
        Grammar gr = new Grammar();
        grammarService.addGrammar(gr);
        Rule nullRule = new Rule();
        Rule nullRightStr = new Rule(null,"dd",null,null);
        Rule emptyStrRule = new Rule(null, "", "", null);
        try {
             ruleService.addRule(nullRule, gr);
            assert false;
        } catch (ResponseStatusException ignored) {

        }
        emptyStrRule = ruleService.addRule(emptyStrRule, gr).getValue0();
        nullRightStr = ruleService.addRule(nullRightStr,gr).getValue0();
        assert emptyStrRule.getLeftSymbol().equals("<empty>");
        assert emptyStrRule.getRightSide().equals("<NULL>");
        assert nullRightStr.getRightSide().equals("<NULL>");
    }

    @Test
    public void testNormalBrackets() {
        grammarService.deleteAll();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "g", "(xd|dd)", false);
        Rule rule1 = new Rule(null, "l", "smile (good|cheer)", false);
        Rule rule2 = new Rule(null, "p", "(pain|bad) cry", false);

        rule = ruleService.addRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals("xd|dd") || rule.getRightSide().equals("dd|xd");
        rule1 = ruleService.addRule(rule1, grammar).getValue0();
        String rule1RightSide = rule1.getRightSide();
        assert rule1RightSide.equals("smile good|smile cheer") || rule1RightSide.equals("smile cheer|smile good");
        rule2 = ruleService.addRule(rule2, grammar).getValue0();
        String rule2RightSide = rule2.getRightSide();
        assert rule2RightSide.equals("pain cry|bad cry") || rule2RightSide.equals("bad cry|pain cry");
        Rule rule3 = new Rule(null, "s", "(a|b) (c|d) (e|f)", false);
        rule3 = ruleService.addRule(rule3, grammar).getValue0();
        String rule3RightSide = rule3.getRightSide();
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        int five = 0;
        int six = 0;
        int seven = 0;
        int eight = 0;
        int nine = 0;
        for (String partString : rule3RightSide.split("\\|")) {
            switch (partString) {
                case "a c e" -> one++;
                case "a c f" -> two++;
                case "a d e" -> three++;
                case "a d f" -> four++;
                case "b c e" -> five++;
                case "b c f" -> six++;
                case "b d e" -> seven++;
                case "b d f" -> eight++;
                default -> nine++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 1 && six == 1 && seven == 1 && eight == 1 && nine == 0;
        Rule rule5 = new Rule(null, "5", "sk(pain|bad)cz", false);
        rule5 = ruleService.addRule(rule5, grammar).getValue0();
        assert rule5.getRightSide().equals("skpaincz|skbadcz") || rule5.getRightSide().equals("skbadcz|skpaincz");
        Rule rule6 = new Rule(null, "6", "((1|2)|(3|4))", false);
        rule6 = ruleService.addRule(rule6, grammar).getValue0();
        one = 0;
        two = 0;
        three = 0;
        four = 0;
        five = 0;
        for (String partString : rule6.getRightSide().split("\\|")) {
            switch (partString) {
                case "1" -> one++;
                case "2" -> two++;
                case "3" -> three++;
                case "4" -> four++;
                default -> five++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 1 && five == 0;

    }


    @Test
    public void testQuoteAndAngleBracket(){
        Grammar grammar =new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "0", "\" xd \"", false);
        rule = ruleService.addRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals("\" xd \"");
        List<String> rightSides= new ArrayList<>();
        List<String> messages= new ArrayList<>();
        rightSides.add("\" xd \"\"");
        messages.add("last quote not ended");
        rightSides.add("<xd><");
        messages.add("ruleref not ended");
        rightSides.add("<xd><dcdc>>");
        messages.add("ending angle bracket without starting bracket");


        for (int i = 0; i < rightSides.size(); i++) {
            Rule errRule = new Rule(null, "0", rightSides.get(i), false);
            try {
                ruleService.addRule(errRule, grammar);
                assert false;
            }catch (ResponseStatusException e){
                assert e.getReason().equals(messages.get(i));
            }
        }

        Rule withAn =new Rule(null, "45", "< xd ><l>", false);
        withAn = ruleService.addRule(withAn, grammar).getValue0();
        assert withAn.getRightSide().equals("< xd ><l>");

        Rule back = new Rule(null, "3", " \" xd \" \\\" ", false);
        back = ruleService.addRule(back, grammar).getValue0();
        assert back.getRightSide().equals("\" xd \" \\\"");

        Rule backNull = new Rule(null, "4", " \" \\\\\"  ", false);
        backNull = ruleService.addRule(backNull, grammar).getValue0();
        assert backNull.getRightSide().equals("\" \\\\\"");

        Rule backAngleBrNull = new Rule(null, "5", " < \\\\>  ", false);
        backAngleBrNull = ruleService.addRule(backAngleBrNull, grammar).getValue0();
        assert backAngleBrNull.getRightSide().equals("< \\\\>");

        Rule backAngleBr = new Rule(null, "6", " < \\> >  ", false);
        backAngleBr = ruleService.addRule(backAngleBr, grammar).getValue0();
        assert backAngleBr.getRightSide().equals("< \\> >");

    }

    @Test
    public void testStraightBrackets() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "0", "[xd]", false);
        rule = ruleService.addRule(rule, grammar).getValue0();
        assert rule.getRightSide().equals("xd|<NULL>") || rule.getRightSide().equals("<NULL>|xd");
        Rule rule1 = new Rule(null, "1", "[a|b]", false);
        rule1 = ruleService.addRule(rule1, grammar).getValue0();
        int one = 0;
        int two = 0;
        int three = 0;
        int four = 0;
        for (String partString : rule1.getRightSide().split("\\|")) {
            switch (partString) {
                case "a" -> one++;
                case "b" -> two++;
                case "<NULL>" -> three++;
                default -> four++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 0;
        one = 0;
        two = 0;
        three = 0;

        Rule rule2 = new Rule(null, "2", "[(a|b) c]", false);
        rule2 = ruleService.addRule(rule2, grammar).getValue0();
        for (String partString : rule2.getRightSide().split("\\|")) {
            switch (partString) {
                case "a c" -> one++;
                case "b c" -> two++;
                case "<NULL>" -> three++;
                default -> four++;
            }
        }
        assert one == 1 && two == 1 && three == 1 && four == 0;
    }

    @Test
    public void testListDeleteAll() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "1", "xd", false);
        Rule rule1 = new Rule(null, "2", "d", true);
        ruleService.addRule(rule, grammar);
        ruleService.addRule(rule1, grammar);
        int one = 0;
        int two = 0;
        int three = 0;
        List<Rule> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        for (Rule partRule : rules) {
            switch (partRule.getLeftSymbol()) {
                case "<1>" -> one++;
                case "<2>" -> two++;
                default -> three++;
            }
        }
        assert one == 1 && two == 1 && three == 0;
        ruleService.deleteRules();
        assert ruleService.getAllRules().size() == 0;
    }


    @Test
    public void testUpdateRule() {
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "old", "xd", false);
        Rule newRule = new Rule(null, "new", "d|d|d", false);
        Rule root = new Rule(null, "root", "l|S|l", true);

        rule = ruleService.addRule(rule, grammar).getValue0();
        root = ruleService.addRule(root, grammar).getValue0();
        Long id = rule.getId();
        newRule = ruleService.updateRule(rule.getId(), newRule);
        assert Objects.equals(id, newRule.getId());
        assert newRule.getRightSide().equals("d");
        assert newRule.getLeftSymbol().equals("<new>");
        root = ruleService.changeRootStatus(root.getId());
        assert !root.getRoot();
        root = ruleService.changeRootStatus(root.getId());
        assert root.getRoot();
        Rule tryNull = new Rule(null, "noch", "noch", false);
        tryNull = ruleService.addRule(tryNull, grammar).getValue0();
        Rule empty = new Rule(null, null, null, null);
        tryNull = ruleService.updateRule(tryNull.getId(), empty);
        assert tryNull.getLeftSymbol().equals("<noch>");
        assert tryNull.getRightSide().equals("noch");
        Rule zeroLen = new Rule(null, "", "", null);
        tryNull = ruleService.updateRule(tryNull.getId(), zeroLen);
        assert tryNull.getLeftSymbol().equals("<noch>");
        assert tryNull.getRightSide().equals("noch");
        Rule xd = new Rule(null, "xd", "sm", null);
        xd = ruleService.addRule(xd, grammar).getValue0();
        xd = ruleService.updateRule(xd.getId(), new Rule(null, null, "1(1|2)", null));
        String result = xd.getRightSide();
        assert !result.contains("(");
    }

    @Test
    public void testFindDeleteById() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Rule rule = new Rule(null, "1", "xd", false);
        Rule rule1 = new Rule(null, "2", "d", true);
        Rule rule3 = new Rule(null, "3", "d", false);
        rule = ruleService.addRule(rule, grammar).getValue0();
        rule1 = ruleService.addRule(rule1, grammar).getValue0();
        ruleService.addRule(rule3, grammar);
        assert ruleService.getById(rule.getId()).getLeftSymbol().equals("<1>");
        ruleService.deleteById(rule1.getId());
        List<Rule> rules = ruleService.getAllRules();
        assert rules.size() == 2;
        int one = 0;
        int def = 0;
        int three = 0;
        for (Rule smallRule : rules) {
            switch (smallRule.getLeftSymbol()) {
                case "<1>" -> one++;
                case "<3>" -> three++;
                default -> def++;
            }
        }
        assert one == 1 && three == 1 && def == 0;


    }

    @Test
    public void testGetByRightSide() {
        ruleService.deleteRules();
        Grammar grammar = new Grammar();
        grammar = grammarService.addGrammar(grammar);
        Grammar grammar2 = new Grammar();
        grammar2 = grammarService.addGrammar(grammar2);
        Rule rule = new Rule(null, "1", "xd", false);
        Rule rule1 = new Rule(null, "2", "d", true);
        Rule rule3 = new Rule(null, "1", "d", false);
        ruleService.addRule(rule, grammar);
        ruleService.addRule(rule1, grammar);
        ruleService.addRule(rule3, grammar2);
        List<Rule> rules = ruleService.getByLeftSide("1");
        assert rules.size() == 2;
        for (Rule smallRule : rules) {
            assert smallRule.getLeftSymbol().equals("<1>");
        }
        List<Rule> rulesFromGrammar = ruleService.getByLeftSide("1", grammar.getId());
        assert rulesFromGrammar.size() == 1;
        for (Rule smallRule : rulesFromGrammar) {
            assert smallRule.getLeftSymbol().equals("<1>") && smallRule.getRightSide().equals("xd");
        }
        assert ruleService.getByLeftSide("5").size() == 0;
    }
}
